from binascii import hexlify

from binary.unsigned_integer.reader import read_bit8

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class MultiSignatureRegistrationDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        self.transaction.asset = {
            "multiSignature": {
                "min": read_bit8(self.serialized, starting_position) & 0xFF,
                "publicKeys": [],
            }
        }

        count = read_bit8(self.serialized, starting_position + 1) & 0xFF

        for index in range(count):
            index_start = int(self.asset_offset) + 4

            if index > 0:
                index_start += index * 66
            public_key = hexlify(self.serialized)[index_start : index_start + 66].decode()
            self.transaction.asset["multiSignature"]["publicKeys"].append(public_key)

        self.transaction.signatures = (
            [] if self.transaction.signatures is None else self.transaction.signatures
        )

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + 4 + (count * 66)
        )

        return self.transaction



from binascii import unhexlify

from solar_crypto.transactions.serializers.base import BaseSerializer


class HtlcRefundSerializer(BaseSerializer):
    """Serializer handling timelock refund data"""

    def serialize(self):
        self.bytes_data += unhexlify(
            self.transaction["asset"]["refund"]["lockTransactionId"].encode()
        )
        return self.bytes_data



from binascii import hexlify, unhexlify

from base58 import b58encode_check
from binary.unsigned_integer.reader import read_bit8, read_bit32, read_bit64

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class HtlcLockDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)
        offset = 0

        self.transaction.amount = read_bit64(self.serialized, offset=starting_position)
        offset += 8

        secret_hash_length = read_bit8(self.serialized, offset=starting_position + 8)
        offset += 1

        secret_hash = hexlify(self.serialized)[
            (starting_position + offset) * 2 : (starting_position + offset + secret_hash_length) * 2
        ]
        offset += secret_hash_length

        expiration_type = read_bit8(self.serialized, offset=starting_position + offset)
        offset += 1

        expiration_value = read_bit32(self.serialized, offset=starting_position + offset)
        offset += 4

        recipient_start_index = (starting_position + offset) * 2
        recipientId = hexlify(self.serialized)[recipient_start_index : recipient_start_index + 42]
        self.transaction.recipientId = b58encode_check(unhexlify(recipientId)).decode()
        offset += 21

        self.transaction.asset["lock"] = {"secretHash": secret_hash.decode()}

        self.transaction.asset["lock"]["expiration"] = {
            "type": expiration_type,
            "value": expiration_value,
        }

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + (offset * 2)
        )

        return self.transaction



from solar_crypto.constants import SOLAR_TRANSACTION_FEES, TRANSACTION_FEES, TRANSACTION_TYPE_GROUP

fees = TRANSACTION_FEES.copy()
solar_fees = SOLAR_TRANSACTION_FEES.copy()


def get_fee(transaction_type, type_group):
    """Get a fee for a given transaction type

    Args:
        transaction_type (int): transaction type for which we wish to get a fee
        type_group (int): transaction type group (TRANSACTION_TYPE_GROUP(Enum))

    Returns:
        int: transaction fee
    """
    if type_group == TRANSACTION_TYPE_GROUP.SOLAR.value:
        return solar_fees.get(transaction_type)
    return fees.get(transaction_type)


def set_fee(transaction_type, type_group, value):
    """Set a fee

    Args:
        transaction_type (int): transaction_type for which we wish to set a fee
        type_group (int): transaction type group (TRANSACTION_TYPE_GROUP(Enum))
        value (int): fee for a given transaction type
    """
    global fees
    global solar_fees

    if type_group == TRANSACTION_TYPE_GROUP.SOLAR.value:
        solar_fees[transaction_type] = value
    else:
        fees[transaction_type] = value



import json

from solar_crypto.identity.private_key import PrivateKey
from solar_crypto.utils.crypto import sign_schnorr, verify_schnorr


class Message(object):
    def __init__(self, **kwargs):
        for k in kwargs.keys():
            if k in ["message", "signature", "publickey", "publicKey"]:
                self.__setattr__(k, kwargs[k])
            else:
                raise TypeError("Invalid keyword argument %s" % k)

    @classmethod
    def sign(cls, message, passphrase):
        """Signs a message

        Args:
            message (str/bytes): a message you wish to sign
            passphrase (str): passphrase you wish to use to sign the message

        Returns:
            Message: returns a message object
        """
        message_bytes = message if isinstance(message, bytes) else message.encode()
        private_key = PrivateKey.from_passphrase(passphrase)
        signature = sign_schnorr(message_bytes, private_key)
        return cls(message=message, signature=signature, publicKey=private_key.public_key)

    def verify(self):
        """Verify the Message object

        Returns:
            bool: returns a boolean - true if verified, false if not
        """
        message = self.message if isinstance(self.message, bytes) else self.message.encode()
        public_key = self.publickey if hasattr(self, "publickey") else self.publicKey
        is_verified = verify_schnorr(message, public_key, self.signature)
        return is_verified

    def to_dict(self):
        """Return a dictionary of the message

        Returns:
            dict: dictionary consiting of public_key, signature and message
        """
        data = {
            ("publicKey" if hasattr(self, "publicKey") else "publickey"): (
                self.publicKey if hasattr(self, "publicKey") else self.publickey
            ),
            "signature": self.signature,
            "message": self.message,
        }
        return data

    def to_json(self):
        """Returns a json string of the the message

        Returns:
            str: json string consisting of public_key, signature and message
        """
        data = self.to_dict()
        return json.dumps(data)



import hashlib
from binascii import unhexlify

from btclib.ecc import ssa

from solar_crypto.identity.private_key import PrivateKey
from solar_crypto.schnorr import schnorr


def sign_schnorr(msg: bytes, private_key: PrivateKey, nonce: int = None) -> str:
    """Signs a message using Schnorr BIP340

    Args:
        msg (bytes): a message you wish to sign
        private_key (PrivateKey): private key object
        nonce (int): deterministic nonce

    Returns:
        str: returnes a hex string of a signature
    """
    signature = ssa.sign(msg, private_key.to_hex(), nonce)
    return signature.serialize().hex()


def sign_schnorr_legacy(msg: bytes, private_key: PrivateKey) -> str:
    """Signs a message using Legacy Schnorr

    Args:
        msg (bytes): a message you wish to sign
        private_key (PrivateKey): private key object
        nonce (int): deterministic nonce

    Returns:
        str: returnes a hex string of a signature
    """
    msg = hashlib.sha256(msg).digest()
    secret = unhexlify(private_key.to_hex())
    signature = schnorr.bcrypto410_sign(msg, secret).hex()
    return signature


def verify_schnorr(msg: bytes, public_key: str, signature: str) -> bool:
    """Verifies a message using Schnorr BIP340

    Args:
        msg (bytes): a message you wish to sign
        public_key (str)
        signature (str)

    Returns:
        bool
    """
    return ssa.verify(msg, public_key, signature)


def verify_schnorr_legacy(msg: bytes, public_key: str, signature: str) -> bool:
    """Verifies a message using Legacy Schnorr

    Args:
        msg (bytes): a message you wish to sign
        public_key (str)
        signature (str)

    Returns:
        bool
    """
    return schnorr.b410_schnorr_verify(msg, public_key, signature)



# From Toons implementation
# see here https://github.com/Moustikitos/dpos/blob/master/dposlib/ark/secp256k1/schnorr.py

import hashlib
from binascii import unhexlify
from builtins import int

p = int(0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F)
n = int(0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141)


def point_from_encoded(pubkey):
    """
    Decode and decompress a ``secp256k1`` point.
    Args:
        pubkey (:class:`bytes`): compressed and encoded point
    Returns:
        :class:`list`: ``secp256k1`` point
    """
    pubkey = bytearray(pubkey)
    x = int_from_bytes(pubkey[1:])
    y = y_from_x(x)
    if y is None:
        raise ValueError("Point not on ``secp256k1`` curve")
    elif y % 2 != pubkey[0] - 2:
        y = -y % p
    return [x, y]


def y_from_x(x):
    """
    Compute :class:`P.y` from :class:`P.x` according to ``y²=x³+7``.
    """
    y_sq = (pow(x, 3, p) + 7) % p
    y = pow(y_sq, (p + 1) // 4, p)
    if pow(y, 2, p) != y_sq:
        return None
    return y


class Point(list):
    """
    ``secp256k1`` point . Initialization can be done with sole ``x`` value.
    :class:`Point` overrides ``*`` and ``+`` operators which accepts
    :class:`list` as argument and returns :class:`Point`. It extends
    :class:`list` and allows item access using ``__getattr__``/``__setattr__``
    operators.
    """

    x = property(
        lambda cls: list.__getitem__(cls, 0),
        lambda cls, v: [
            list.__setitem__(cls, 0, int(v)),
            list.__setitem__(cls, 1, y_from_x(int(v))),
        ],
        None,
        "",
    )
    y = property(lambda cls: list.__getitem__(cls, 1), None, None, "")

    def __init__(self, *xy):
        if len(xy) == 0:
            xy = (0, None)
        elif len(xy) == 1:
            xy += (y_from_x(int(xy[0])),)
        list.__init__(self, [int(e) if e is not None else e for e in xy[:2]])

    def __mul__(self, k):
        if isinstance(k, int):
            return Point(*point_mul(self, k))
        else:
            raise TypeError("'%s' should be an int" % k)

    __rmul__ = __mul__

    def __add__(self, P):
        if isinstance(P, list):
            return Point(*point_add(self, P))
        else:
            raise TypeError("'%s' should be a 2-int-length list" % P)

    __radd__ = __add__

    @staticmethod
    def decode(pubkey):
        """
        Decode and decompress a ``secp256k1`` point.
        """
        return Point(*point_from_encoded(pubkey))

    def encode(self):
        """
        Decode and decompress a ``secp256k1`` point.
        """
        return encoded_from_point(self)


G = Point(0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798)


def point_add(P1, P2):
    """
    Add ``secp256k1`` points.
    Args:
        P1 (:class:`list`): first ``secp256k1`` point
        P2 (:class:`list`): second ``secp256k1`` point
    Returns:
        :class:`list`: ``secp256k1`` point
    """
    if P1 is None:
        return P2
    if P2 is None:
        return P1
    if x(P1) == x(P2) and y(P1) != y(P2):
        raise ValueError("One of the point is not on the curve")
    if P1 == P2:
        lam = (3 * x(P1) * x(P1) * pow(2 * y(P1), p - 2, p)) % p
    else:
        lam = ((y(P2) - y(P1)) * pow(x(P2) - x(P1), p - 2, p)) % p
    x3 = (lam * lam - x(P1) - x(P2)) % p
    return [x3, (lam * (x(P1) - x3) - y(P1)) % p]


def bcrypto410_verify(msg, pubkey, sig):
    if len(msg) != 32:
        raise ValueError("The message must be a 32-byte array.")
    if len(sig) != 64:
        raise ValueError("The signature must be a 64-byte array.")

    P = Point.decode(pubkey)
    r, s = int_from_bytes(sig[:32]), int_from_bytes(sig[32:])
    if r >= p or s >= n:
        return False

    e = int_from_bytes(hash_sha256(sig[0:32] + pubkey + msg)) % n
    R = Point(*(G * s + point_mul(P, n - e)))  # P*(n-e) does not work...
    if R is None or not is_quad(R.y) or R.x != r:
        return False
    return True


def b410_schnorr_verify(message, publicKey, signature):
    return bcrypto410_verify(
        hash_sha256(message), Point.decode(unhexlify(publicKey)).encode(), unhexlify(signature)
    )


def x(P):
    """
    Return :class:`P.x` or :class:`P[0]`.
    Args:
        P (:class:`list`): ``secp256k1`` point
    Returns:
        :class:`int`: x
    """
    return P[0]


def bytes_from_int(x):
    return int(x).to_bytes(32, byteorder="big")


def int_from_bytes(b):
    return int.from_bytes(b, byteorder="big")


def jacobi(x):
    return pow(x, (p - 1) // 2, p)


def is_quad(x):
    return jacobi(x) == 1


def hash_sha256(b):
    """
    Args:
        b (:class:`bytes` or :class:`str`): sequence to be hashed
    Returns:
        :class:`bytes`: sha256 hash
    """
    return hashlib.sha256(b if isinstance(b, bytes) else b.encode("utf-8")).digest()


def point_mul(P, n):
    """
    Multiply ``secp256k1`` point with scalar.
    Args:
        P (:class:`list`): ``secp256k1`` point
        n (:class:`int`): scalar
    Returns:
        :class:`list`: ``secp256k1`` point
    """
    R = None
    for i in range(256):
        if (n >> i) & 1:
            R = point_add(R, P)
        P = point_add(P, P)
    return R


def y(P):
    """
    Return :class:`P.y` or :class:`P[1]`.

    Args:
        P (:class:`list`): ``secp256k1`` point
    Returns:
        :class:`int`: y
    """
    return P[1]


def encoded_from_point(P):
    """
    Encode and compress a ``secp256k1`` point:
      * ``bytes(2) || bytes(x)`` if y is even
      * ``bytes(3) || bytes(x)`` if y is odd

    Args:
        P (:class:`list`): ``secp256k1`` point
    Returns:
        :class:`bytes`: compressed and encoded point
    """
    return (b"\x03" if y(P) & 1 else b"\x02") + bytes_from_int(x(P))


# https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js
def bcrypto410_sign(msg, seckey0):
    if len(msg) != 32:
        raise ValueError("The message must be a 32-byte array.")

    seckey = int_from_bytes(seckey0)
    if not (1 <= seckey <= n - 1):
        raise ValueError("The secret key must be an integer in the range 1..n-1.")

    k0 = int_from_bytes(hash_sha256(seckey0 + msg)) % n
    if k0 == 0:
        raise RuntimeError("Failure. This happens only with negligible probability.")

    R = G * k0
    Rraw = bytes_from_int(R.x)
    e = int_from_bytes(hash_sha256(Rraw + encoded_from_point(G * seckey) + msg)) % n

    seckey %= n
    k0 %= n
    k = n - k0 if not is_quad(R.y) else k0

    s = (k + e * seckey) % n
    s %= n

    return Rraw + bytes_from_int(s)



from enum import Enum

TRANSACTION_TRANSFER = 0
TRANSACTION_SECOND_SIGNATURE_REGISTRATION = 1
TRANSACTION_DELEGATE_REGISTRATION = 2
TRANSACTION_VOTE = 3
TRANSACTION_MULTI_SIGNATURE_REGISTRATION = 4
TRANSACTION_IPFS = 5
TRANSACTION_MULTI_PAYMENT = 6
TRANSACTION_DELEGATE_RESIGNATION = 7
TRANSACTION_HTLC_LOCK = 8
TRANSACTION_HTLC_CLAIM = 9
TRANSACTION_HTLC_REFUND = 10

TRANSACTION_TYPES = {
    TRANSACTION_TRANSFER: "transfer",
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION: "second_signature_registration",
    TRANSACTION_DELEGATE_REGISTRATION: "delegate_registration",
    TRANSACTION_VOTE: "vote",
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION: "multi_signature_registration",
    TRANSACTION_IPFS: "ipfs",
    TRANSACTION_MULTI_PAYMENT: "multi_payment",
    TRANSACTION_DELEGATE_RESIGNATION: "delegate_resignation",
    TRANSACTION_HTLC_LOCK: "htlc_lock",
    TRANSACTION_HTLC_CLAIM: "htlc_claim",
    TRANSACTION_HTLC_REFUND: "htlc_refund",
}

TRANSACTION_FEES = {
    TRANSACTION_DELEGATE_REGISTRATION: 7500000000,
    TRANSACTION_TRANSFER: 5000000,
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION: 5000000,
    TRANSACTION_VOTE: 5000000,
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION: 5000000,
    TRANSACTION_IPFS: 5000000,
    TRANSACTION_MULTI_PAYMENT: 50000000,
    TRANSACTION_HTLC_LOCK: 5000000,
    TRANSACTION_DELEGATE_RESIGNATION: 0,
    TRANSACTION_HTLC_CLAIM: 0,
    TRANSACTION_HTLC_REFUND: 0,
}

SOLAR_TRANSACTION_BURN = 0

SOLAR_TRANSACTION_TYPES = {
    SOLAR_TRANSACTION_BURN: "burn",
}

SOLAR_TRANSACTION_FEES = {
    SOLAR_TRANSACTION_BURN: 0,
}


class TRANSACTION_TYPE_GROUP(Enum):
    def __int__(self):
        return int(self.value)

    TEST = 0
    CORE = 1
    SOLAR = 2
    RESERVED = 1000  # Everything above is available to anyone


class HTLC_LOCK_EXPIRATION_TYPE(Enum):
    def __str__(self):
        return int(self.value)

    EPOCH_TIMESTAMP = 1
    BLOCK_HEIGHT = 2


class HashingType(Enum):
    SHA256 = 0
    SHA384 = 1
    SHA512 = 2
    SHA3256 = 3
    SHA3384 = 4
    SHA3512 = 5
    Keccak256 = 6
    Keccak384 = 7
    Keccak512 = 8



from binascii import hexlify

from binary.unsigned_integer.reader import read_bit8

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class HtlcClaimDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)
        offset = 0

        hash_type = read_bit8(self.serialized, offset=starting_position)
        offset += 1

        lock_transaction_id = hexlify(
            self.serialized[starting_position + offset : starting_position + offset + 32]
        )
        offset += 32

        unlock_secret_length = read_bit8(self.serialized, offset=starting_position + offset)
        offset += 1

        unlock_secret = hexlify(
            self.serialized[
                starting_position + offset : starting_position + offset + unlock_secret_length
            ]
        )
        offset += unlock_secret_length

        self.transaction.asset["claim"] = {
            "hashType": hash_type,
            "lockTransactionId": lock_transaction_id.decode(),
            "unlockSecret": unlock_secret.decode(),
        }

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + (offset * 2)
        )

        return self.transaction



from binary.unsigned_integer.writer import write_bit64

from solar_crypto.transactions.serializers.base import BaseSerializer


class BurnSerializer(BaseSerializer):
    def serialize(self):
        self.bytes_data += write_bit64(self.transaction["amount"])
        return self.bytes_data



from solar_crypto.constants import TRANSACTION_HTLC_LOCK, TRANSACTION_TYPE_GROUP
from solar_crypto.identity import address
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class HtlcLock(BaseTransactionBuilder):

    transaction_type = TRANSACTION_HTLC_LOCK

    def __init__(
        self,
        recipient_id,
        amount,
        secret_hash,
        expiration_type,
        expiration_value,
        vendorField=None,
        fee=None,
    ):
        """Create a timelock transaction

        Args:
            recipient_id (str): recipient identifier
            amount (int): amount of coins you want to transfer
            secret_hash (str): a hash of the secret. The SAME hash must be used in the corresponding “claim” transaction
            expiration_type (int): type of the expiration. Either block height or network epoch timestamp based
            expiration_value (int): Expiration of transaction in seconds or height depending on expiration_type
            vendorField (str): value for the vendor field aka smartbridge
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        if not address.validate_address(recipient_id):
            raise ValueError("Invalid recipient address")

        self.transaction.recipientId = recipient_id

        if type(amount) == int and amount > 0:
            self.transaction.amount = amount
        else:
            raise ValueError("Amount is not valid")

        self.transaction.typeGroup = self.get_type_group()

        self.transaction.asset["lock"] = {
            "secretHash": secret_hash,
            "expiration": {"type": expiration_type, "value": expiration_value},
        }

        self.transaction.vendorField = vendorField.encode() if vendorField else None

        if fee:
            self.transaction.fee = fee

    def get_type_group(self):
        return TRANSACTION_TYPE_GROUP.CORE.value



from solar_crypto.constants import TRANSACTION_HTLC_REFUND, TRANSACTION_TYPE_GROUP
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class HtlcRefund(BaseTransactionBuilder):

    transaction_type = TRANSACTION_HTLC_REFUND

    def __init__(self, lock_transaction_id, fee=None):
        """Create a timelock transaction

        Args:
            lock_transaction_id (str) : HTLC lock transaction ID we wish to claim
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        self.transaction.typeGroup = self.get_type_group()

        self.transaction.asset["refund"] = {
            "lockTransactionId": lock_transaction_id,
        }

        if fee:
            self.transaction.fee = fee

    def get_type_group(self):
        return TRANSACTION_TYPE_GROUP.CORE.value



from binascii import hexlify

from base58 import b58decode
from binary.hex.writer import write_high

from solar_crypto.transactions.serializers.base import BaseSerializer


class IPFSSerializer(BaseSerializer):
    """Serializer handling ipfs data"""

    def serialize(self):
        ipfs = hexlify(b58decode(self.transaction["asset"]["ipfs"]))

        self.bytes_data += write_high(ipfs)

        return self.bytes_data



from binascii import hexlify

from binary.unsigned_integer.reader import read_bit8

from solar_crypto.exceptions import SolarDeserializerException
from solar_crypto.transactions.deserializers.base import BaseDeserializer


class VoteDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)
        offset = 0

        vote_length = read_bit8(self.serialized, offset=starting_position)
        offset += 1

        self.transaction.asset["votes"] = []

        for _ in range(vote_length):
            if (
                self.transaction.version == 2
                and self.serialized[starting_position + offset : starting_position + offset + 1]
                != b"\xff"
            ):
                vote_buffer = self.serialized[
                    starting_position + offset : starting_position + offset + 34
                ]
                offset += 34
                prefix = "+" if vote_buffer[0] == 1 else "-"
                vote = f"{prefix}{vote_buffer[1::].hex()}"
            else:
                if self.transaction.version == 2:
                    offset += 1  # +1 due to NOT moving forwards when checking for `b"\xff"`

                length = read_bit8(
                    self.serialized[starting_position + offset : starting_position + offset + 1]
                )
                offset += 1

                vote_buffer = self.serialized[
                    starting_position + offset : starting_position + offset + length
                ]
                offset += length

                prefix = "+" if vote_buffer[0] == 1 else "-"
                vote = f"{prefix}{vote_buffer[1::].decode()}"

            if len(vote) <= 1:
                raise SolarDeserializerException("Invalid transaction data")

            self.transaction.asset["votes"].append(vote)

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + 2 + ((offset - 1) * 2)
        )

        return self.transaction



from solar_crypto.constants import TRANSACTION_MULTI_PAYMENT, TRANSACTION_TYPE_GROUP
from solar_crypto.identity import address
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class MultiPayment(BaseTransactionBuilder):

    transaction_type = TRANSACTION_MULTI_PAYMENT

    def __init__(self, vendorField=None, fee=None):
        """Create a multi payment transaction

        Args:
            vendorField (str): value for the vendor field aka smartbridge
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        self.transaction.typeGroup = self.get_type_group()

        self.transaction.asset["payments"] = []

        self.transaction.vendorField = vendorField.encode() if vendorField else None

        if fee:
            self.transaction.fee = fee

    def get_type_group(self):
        return TRANSACTION_TYPE_GROUP.CORE.value

    def add_payment(self, amount, recipient_id):
        if not address.validate_address(recipient_id):
            raise ValueError("Invalid recipient address")

        self.transaction.asset["payments"].append({"amount": amount, "recipientId": recipient_id})



from binascii import hexlify, unhexlify

from base58 import b58encode
from binary.unsigned_integer.reader import read_bit8
from binary.unsigned_integer.writer import write_bit8

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class IPFSDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        hash_function = read_bit8(self.serialized, starting_position) & 0xFF
        ipfs_hash_length = read_bit8(self.serialized, starting_position + 1) & 0xFF

        start_index = (starting_position + 2) * 2
        end_index = start_index + (ipfs_hash_length * 2)
        ipfs_hash = hexlify(self.serialized)[start_index:end_index]

        temp_buffer = bytes()

        temp_buffer += write_bit8(hash_function)
        temp_buffer += write_bit8(ipfs_hash_length)
        temp_buffer += unhexlify(ipfs_hash)

        self.transaction.asset = {"ipfs": b58encode(temp_buffer).decode()}

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + (ipfs_hash_length + 2) * 2
        )

        return self.transaction



from binascii import hexlify

from binary.unsigned_integer.reader import read_bit64

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class BurnDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)
        self.transaction.amount = read_bit64(self.serialized, offset=starting_position)
        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(), self.asset_offset + 8 * 2
        )
        return self.transaction



from solar_crypto.configuration.fee import get_fee
from solar_crypto.constants import HTLC_LOCK_EXPIRATION_TYPE, TRANSACTION_TYPE_GROUP
from solar_crypto.identity.private_key import PrivateKey
from solar_crypto.identity.public_key import PublicKey
from solar_crypto.transactions.transaction import Transaction
from solar_crypto.utils.crypto import sign_schnorr, sign_schnorr_legacy


class BaseTransactionBuilder(object):
    def __init__(self):
        self.transaction = Transaction()
        self.transaction.type = getattr(self, "transaction_type", None)
        self.transaction.fee = get_fee(
            getattr(self, "transaction_type", None), getattr(self, "typeGroup", 1)
        )
        self.transaction.nonce = getattr(self, "nonce", None)
        self.transaction.typeGroup = getattr(self, "typeGroup", 1)
        self.transaction.signatures = getattr(self, "signatures", None)
        self.transaction.version = getattr(self, "version", 3)
        if self.transaction.type != 0:
            self.transaction.amount = getattr(self, "amount", 0)

    def to_dict(self):
        return self.transaction.to_dict()

    def to_json(self):
        return self.transaction.to_json()

    def sign(self, passphrase):
        """Sign the transaction using the given passphrase

        Args:
            passphrase (str): passphrase associated with the account sending this transaction
        """
        pvt = PrivateKey.from_passphrase(passphrase)
        self.transaction.senderPublicKey = PublicKey.from_passphrase(passphrase)
        msg = self.transaction.to_bytes(True, True, False)

        if self.transaction.version > 2:
            # schnorr bip340
            sig = sign_schnorr(msg, pvt)
            self.transaction.signature = sig
            self.transaction.id = self.transaction.get_id()
        else:
            # schnorr legacy
            sig = sign_schnorr_legacy(msg, pvt)
            self.transaction.signature = sig
            self.transaction.id = self.transaction.get_id()

    def second_sign(self, passphrase):
        """Sign the transaction using the given second passphrase

        Args:
            passphrase (str): 2nd passphrase associated with the account sending this transaction
        """
        pvt = PrivateKey.from_passphrase(passphrase)
        msg = self.transaction.to_bytes(False, True, False)

        if self.transaction.version > 2:
            # schnorr bip340
            sig = sign_schnorr(msg, pvt)
            self.transaction.signSignature = sig
            self.transaction.id = self.transaction.get_id()
        else:
            # schnorr legacy
            sig = sign_schnorr_legacy(msg, pvt)
            self.transaction.signSignature = sig
            self.transaction.id = self.transaction.get_id()

    def multi_sign(self, passphrase, index):
        if not self.transaction.signatures:
            self.transaction.signatures = []

        index = len(self.transaction.signatures) if index == -1 else index
        pvt = PrivateKey.from_passphrase(passphrase)
        index_formatted = hex(index).replace("x", "")
        msg = self.transaction.to_bytes(True, True, True)

        if self.transaction.version > 2:
            # schnorr bip340
            sig = sign_schnorr(msg, pvt)
            indexed_signature = f"{index_formatted}{sig}"
            self.transaction.signatures.append(indexed_signature)
        else:
            # schnorr legacy
            sig = sign_schnorr_legacy(msg, pvt)
            self.transaction.signatures.append(f"{index_formatted}{sig}")

    def verify(self):
        return self.transaction.verify()

    def verify_second(self, secondPublicKey):
        return self.transaction.verify_secondsig(secondPublicKey)

    def verify_multisig(self, multi_signature_asset):
        return self.transaction.verify_signatures(multi_signature_asset)

    def set_nonce(self, nonce):
        self.transaction.nonce = nonce

    def set_fee(self, fee: int):
        self.transaction.fee = fee

    def set_amount(self, amount):
        self.transaction.amount = amount

    def set_sender_public_key(self, public_key):
        self.transaction.senderPublicKey = public_key

    def set_expiration(self, expiration):
        if type(expiration) == int:
            self.transaction.expiration = expiration
        else:
            types = {
                HTLC_LOCK_EXPIRATION_TYPE.EPOCH_TIMESTAMP: 1,
                HTLC_LOCK_EXPIRATION_TYPE.BLOCK_HEIGHT: 2,
            }
            self.transaction.expiration = types[expiration]

    def set_type_group(self, type_group):
        if type(type_group) == int:
            self.transaction.typeGroup = type_group
        else:
            types = {
                TRANSACTION_TYPE_GROUP.TEST: 0,
                TRANSACTION_TYPE_GROUP.CORE: 1,
                TRANSACTION_TYPE_GROUP.RESERVED: 1000,
            }
            self.transaction.typeGroup = types[type_group]

    def set_version(self, version):
        self.transaction.version = version

    def set_vendor_field(self, value: str):
        self.transaction.vendorField = value.encode()



from binascii import hexlify, unhexlify

from binary.unsigned_integer.writer import write_bit8

from solar_crypto.transactions.serializers.base import BaseSerializer


class VoteSerializer(BaseSerializer):
    """Serializer handling vote data"""

    def serialize(self):
        vote_bytes = []

        for vote in self.transaction["asset"]["votes"]:
            prefix = "01" if vote.startswith("+") else "00"
            sliced = vote[1::]

            if len(sliced) == 66:
                vote_bytes.append(f"{prefix}{sliced}")
                continue

            # vote.length.toString(16).padStart(2, "0") + prefix + Buffer.from(sliced).toString("hex");
            start = format(len(vote), "x").zfill(2)
            vote_hex = f"{start}{prefix}{hexlify(sliced.encode()).decode()}"
            if self.transaction["version"] == 2:
                vote_hex = f"ff{vote_hex}"

            vote_bytes.append(vote_hex)

        self.bytes_data += write_bit8(len(self.transaction["asset"]["votes"]))
        self.bytes_data += unhexlify("".join(vote_bytes))

        return self.bytes_data



from binascii import unhexlify

from binary.unsigned_integer.writer import write_bit8

from solar_crypto.transactions.serializers.base import BaseSerializer


class MultiSignatureSerializer(BaseSerializer):
    """Serializer handling delegate registration data"""

    def serialize(self):
        public_keys_length = len(self.transaction["asset"]["multiSignature"]["publicKeys"])
        self.bytes_data += write_bit8(self.transaction["asset"]["multiSignature"]["min"])
        self.bytes_data += write_bit8(public_keys_length)
        for key in self.transaction["asset"]["multiSignature"]["publicKeys"]:
            self.bytes_data += unhexlify(key.encode())

        return self.bytes_data



import inspect
from binascii import hexlify, unhexlify
from importlib import import_module

from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit8, write_bit16, write_bit32, write_bit64

from solar_crypto.configuration.network import get_network_version
from solar_crypto.constants import (
    SOLAR_TRANSACTION_TYPES,
    TRANSACTION_TYPE_GROUP,
    TRANSACTION_TYPES,
)
from solar_crypto.exceptions import SolarSerializerException
from solar_crypto.transactions.serializers.base import BaseSerializer


class Serializer(object):

    transaction = None

    def __init__(self, transaction):
        if not transaction:
            raise SolarSerializerException("No transaction data provided")
        self.transaction = transaction

    def serialize(
        self, skip_signature=True, skip_second_signature=True, skip_multi_signature=True, raw=False
    ):
        """Perform AIP11 compliant serialization

        Returns:
            bytes: bytes string
        """
        bytes_data = bytes()

        bytes_data += write_bit8(0xFF)
        bytes_data += write_bit8(self.transaction.get("version") or 0x02)
        bytes_data += write_bit8(self.transaction.get("network") or get_network_version())
        bytes_data += write_bit32(self.transaction.get("typeGroup") or 0x01)
        bytes_data += write_bit16(self.transaction.get("type"))
        bytes_data += write_bit64(self.transaction.get("nonce") or 0x01)

        bytes_data += write_high(self.transaction.get("senderPublicKey"))
        bytes_data += write_bit64(self.transaction.get("fee"))

        if self.transaction.get("vendorField"):
            vendorFieldLength = len(self.transaction.get("vendorField"))
            bytes_data += write_bit8(vendorFieldLength)
            bytes_data += self.transaction["vendorField"].encode()
        else:
            bytes_data += write_bit8(0x00)

        bytes_data = self._handle_transaction_type(bytes_data)
        bytes_data = self._handle_signature(
            bytes_data, skip_signature, skip_second_signature, skip_multi_signature
        )

        return bytes_data if raw else hexlify(bytes_data).decode()

    def _handle_transaction_type(self, bytes_data):
        """Serialize transaction specific data (eg. delegate registration)

        Args:
            bytes_data (bytes): already serialized data about a transaction (eg. version, network)

        Returns:
            bytes: bytes string
        """
        serializer_name = self._get_serializer_name()
        module = import_module("solar_crypto.transactions.serializers.{}".format(serializer_name))
        for attr in dir(module):
            # If attr name is `BaseSerializer`, skip it as it's a class and also has a
            # subclass of BaseSerializer
            if attr == "BaseSerializer":
                continue

            attribute = getattr(module, attr)
            if inspect.isclass(attribute) and issubclass(attribute, BaseSerializer):
                # this attribute is actually a specific serializer that we want to use
                serializer = attribute
                break
        return serializer(self.transaction, bytes_data).serialize()

    def _handle_signature(
        self, bytes_data, skip_signature, skip_second_signature, skip_multi_signature
    ):
        """Serialize signature data of the transaction

        Args:
            bytes_data (bytes): already serialized data

        Returns:
            bytes: bytes string
        """
        if not skip_signature and self.transaction.get("signature"):
            bytes_data += unhexlify(self.transaction["signature"])

        if not skip_second_signature and self.transaction.get("secondSignature"):
            bytes_data += unhexlify(self.transaction["secondSignature"])
        if not skip_second_signature and self.transaction.get("signSignature"):
            bytes_data += unhexlify(self.transaction["signSignature"])
        if not skip_multi_signature and self.transaction.get("signatures"):
            bytes_data += unhexlify("".join(self.transaction["signatures"]))

        return bytes_data

    def _get_serializer_name(self):
        if self.transaction["typeGroup"] == TRANSACTION_TYPE_GROUP.SOLAR.value:
            name = SOLAR_TRANSACTION_TYPES[self.transaction["type"]]
        else:
            name = TRANSACTION_TYPES[self.transaction["type"]]
        return name



from binascii import hexlify

from base58 import b58decode_check
from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit16, write_bit64

from solar_crypto.exceptions import SolarSerializerException
from solar_crypto.identity import address
from solar_crypto.transactions.serializers.base import BaseSerializer


class MultiPaymentSerializer(BaseSerializer):
    """Serializer handling multi payment data"""

    def serialize(self):
        self.bytes_data += write_bit16(len(self.transaction["asset"]["payments"]))

        for payment in self.transaction["asset"]["payments"]:
            if not address.validate_address(payment["recipientId"]):
                raise SolarSerializerException("Invalid recipient address")

            self.bytes_data += write_bit64(payment["amount"])
            recipientId = hexlify(b58decode_check(payment["recipientId"]))
            self.bytes_data += write_high(recipientId)

        return self.bytes_data



from solar_crypto.networks.testnet import Testnet

network = {}


def set_network(network_object):
    """Set what network you want to use in the crypto library

    Args:
        network_object (Network object): Testnet, Mainnet
    """
    global network
    network = {
        "epoch": network_object.epoch,
        "version": network_object.version,
        "wif": network_object.wif,
    }


def get_network():
    """Get settings for a selected network, default network is testnet

    Returns:
        dict: network settings (default network is testnet)
    """
    if not network:
        set_network(Testnet)
    return network


def set_custom_network(epoch, version, wif):
    """Set custom network

    Args:
        epoch (datetime): chains epoch time
        version (int): chains version
        wif (int): chains wif
    """
    global network
    network = {
        "epoch": epoch,
        "version": version,
        "wif": wif,
    }


def get_network_version():
    """Get currently set network version

    Returns:
        int: network version
    """
    network = get_network()
    return network["version"]



from binascii import hexlify

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class HtlcRefundDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset)

        lock_transaction_id = hexlify(self.serialized)[starting_position : starting_position + 64]

        self.transaction.asset["refund"] = {"lockTransactionId": lock_transaction_id.decode()}

        self.transaction.parse_signatures(hexlify(self.serialized).decode(), self.asset_offset + 64)

        return self.transaction



import inspect
from binascii import hexlify, unhexlify
from importlib import import_module

from binary.unsigned_integer.reader import read_bit8, read_bit16, read_bit32, read_bit64

from solar_crypto.constants import (
    SOLAR_TRANSACTION_TYPES,
    TRANSACTION_TYPE_GROUP,
    TRANSACTION_TYPES,
)
from solar_crypto.exceptions import SolarDeserializerException, SolarSerializerException
from solar_crypto.identity import address
from solar_crypto.transactions.deserializers.base import BaseDeserializer


class Deserializer(object):

    serialized = None

    def __init__(self, serialized):
        self.serialized = unhexlify(serialized)

    def deserialize(self):
        """Deserialize transaction

        Returns:
            object: returns Transaction resource object
        """
        # circular import with transaction.py :( - I'm thinking of just returning a dict here
        # which then needs to be passed to a Transaction object, instead of returning a Transaction
        # object
        from solar_crypto.transactions.transaction import Transaction

        transaction = Transaction()
        transaction.version = read_bit8(self.serialized, offset=1)
        transaction.network = read_bit8(self.serialized, offset=2)
        transaction.typeGroup = read_bit32(self.serialized, offset=3)
        transaction.type = read_bit16(self.serialized, offset=7)
        transaction.nonce = read_bit64(self.serialized, offset=9)
        transaction.senderPublicKey = hexlify(self.serialized)[34 : 66 + 34].decode()
        transaction.fee = read_bit64(self.serialized, offset=50)

        vendor_field_length = read_bit8(self.serialized, offset=58)
        if vendor_field_length > 0:
            vendor_field_offset = 59
            vendorField_end = vendor_field_offset + vendor_field_length
            transaction.vendorField = self.serialized[vendor_field_offset:vendorField_end].decode()

        asset_offset = (58 + 1) * 2 + vendor_field_length * 2
        handled_transaction = self._handle_transaction_type(asset_offset, transaction)
        transaction.amount = handled_transaction.amount
        transaction.version = handled_transaction.version

        if transaction.version not in [2, 3]:
            raise SolarDeserializerException(
                "%s is not a valid transaction version", transaction.version
            )

        self._validate(transaction)

        return transaction

    def _handle_transaction_type(self, asset_offset, transaction):
        """Handle deserialization for a given transaction type

        Args:
            asset_offset (int):
            transaction (object): Transaction resource object

        Returns:
            object: Transaction resource object of currently deserialized data
        """
        deserializer_name = self._get_deserializer_name(transaction)
        module = import_module(
            "solar_crypto.transactions.deserializers.{}".format(deserializer_name)
        )
        for attr in dir(module):
            # If attr name is `BaseDeserializer`, skip it as it's a class and also has a
            # subclass of BaseDeserializer
            if attr == "BaseDeserializer":
                continue

            attribute = getattr(module, attr)
            if inspect.isclass(attribute) and issubclass(attribute, BaseDeserializer):
                # this attribute is actually a specific deserializer that we want to use
                deserializer = attribute
                break
        return deserializer(self.serialized, asset_offset, transaction).deserialize()

    @staticmethod
    def _get_deserializer_name(transaction):
        if transaction.typeGroup == TRANSACTION_TYPE_GROUP.SOLAR.value:
            name = SOLAR_TRANSACTION_TYPES[transaction.type]
        else:
            name = TRANSACTION_TYPES[transaction.type]
        return name

    @staticmethod
    def _validate(transaction):
        if not transaction.recipientId:
            return

        if not address.validate_address(transaction.recipientId):
            raise SolarSerializerException("Invalid recipient address")



from binascii import hexlify

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class DelegateResignationDeserializer(BaseDeserializer):
    def deserialize(self):
        self.transaction.parse_signatures(hexlify(self.serialized).decode(), self.asset_offset)

        return self.transaction



import json
from binascii import unhexlify
from hashlib import sha256

from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit8

from solar_crypto.configuration.network import get_network_version
from solar_crypto.constants import (
    TRANSACTION_DELEGATE_REGISTRATION,
    TRANSACTION_MULTI_SIGNATURE_REGISTRATION,
    TRANSACTION_SECOND_SIGNATURE_REGISTRATION,
    TRANSACTION_VOTE,
)
from solar_crypto.exceptions import SolarInvalidTransaction
from solar_crypto.transactions.deserializer import Deserializer
from solar_crypto.transactions.serializer import Serializer
from solar_crypto.utils.crypto import verify_schnorr, verify_schnorr_legacy

TRANSACTION_ATTRIBUTES = {
    "amount": 0,
    "asset": dict,
    "fee": None,
    "id": None,
    "network": get_network_version,
    "recipientId": None,
    "secondSignature": None,
    "senderPublicKey": None,
    "signature": None,
    "signatures": None,
    "signSignature": None,
    "nonce": None,
    "type": None,
    "typeGroup": None,
    "vendorField": None,
    "version": None,
    "lockTransactionId": None,
    "lockSecret": None,
    "expiration": None,
}


class Transaction(object):
    def __init__(self, *args, **kwargs):
        for attribute, attribute_value in TRANSACTION_ATTRIBUTES.items():
            if callable(attribute_value):
                attribute_value = attribute_value()
            if attribute in kwargs:
                attribute_value = kwargs[attribute]
            setattr(self, attribute, attribute_value)

    def get_id(self):
        """Convert the byte representation to a unique identifier

        Returns:
            str:
        """
        return sha256(
            self.to_bytes(
                skip_signature=False, skip_second_signature=False, skip_multi_signature=False
            )
        ).hexdigest()

    def to_dict(self):
        """Convert the transaction into a dictionary representation

        Returns:
            dict: only includes values that are set
        """
        data = {}
        for key in TRANSACTION_ATTRIBUTES.keys():
            attribute = getattr(self, key, None)
            if attribute is None:
                continue
            # todo: get rid of the bytes check and handle this outside of the to_dict function
            data[key] = attribute.decode() if isinstance(attribute, bytes) else attribute
        return data

    def to_json(self):
        data = self.to_dict()
        return json.dumps(data)

    def to_bytes(self, skip_signature=True, skip_second_signature=True, skip_multi_signature=True):
        """Convert the transaction to its byte representation

        Args:
            skip_signature (bool, optional): do you want to skip the signature
            skip_second_signature (bool, optional): do you want to skip the 2nd signature
            skip_multi_signature (bool, optional): do you want to skip multi signature

        Returns:
            bytes: bytes representation of the transaction
        """
        return Serializer(self.to_dict()).serialize(
            skip_signature=skip_signature,
            skip_second_signature=skip_second_signature,
            skip_multi_signature=skip_multi_signature,
            raw=True,
        )

    def parse_signatures(self, serialized, start_offset):
        """Parse the signature, second signature and multi signatures

        Args:
            serialized (str): parses a given serialized string
            start_offset (int):

        Returns:
            None: methods returns nothing
        """

        signature_end_offset = start_offset + (64 * 2)

        if len(serialized) - signature_end_offset % 65 != 0:
            self.signature = serialized[start_offset:signature_end_offset]

        second_signature_end_offset = signature_end_offset + (64 * 2)
        if (
            len(serialized) - signature_end_offset > 0
            and (len(serialized) - signature_end_offset) % 64 == 0
        ):
            self.signSignature = serialized[signature_end_offset:second_signature_end_offset]

        if (
            len(serialized) - second_signature_end_offset > 0
            and (len(serialized) - signature_end_offset) % 65 == 0
        ):
            multi_sig_part = serialized[signature_end_offset:]
            index = 0
            index_size = 2
            signature_size = 128

            while index != len(multi_sig_part):
                signature_index = multi_sig_part[index : index + index_size]
                signature = multi_sig_part[index + index_size : index + index_size + signature_size]
                index += index_size + signature_size
                signature_formatted = signature_index + signature
                self.signatures.append(signature_formatted)

        return

    def serialize(self, skip_signature=True, skip_second_signature=True, skip_multi_signature=True):
        """Perform AIP11 compliant serialization.

        Args:
            skip_signature (bool, optional): do you want to skip the signature
            skip_second_signature (bool, optional): do you want to skip the 2nd signature
            skip_multi_signature (bool, optional): do you want to skip multi signature

        Returns:
            str: Serialized string
        """
        data = self.to_dict()
        return Serializer(data).serialize(
            skip_signature, skip_second_signature, skip_multi_signature
        )

    def deserialize(self, serialized):
        """Perform AIP11 compliant deserialization.

        Args:
            serialized (str): parses a given serialized string

        Returns:
            crypto.transactions.transaction.Transaction: Transaction
        """
        return Deserializer(serialized).deserialize()

    def verify(self):
        msg = self.to_bytes()

        if self.version > 2:
            # schnorr bip340
            is_valid = verify_schnorr(msg, self.senderPublicKey, self.signature)
        else:
            # schnorr legacy
            is_valid = verify_schnorr_legacy(msg, self.senderPublicKey, self.signature)

        if not is_valid:
            raise SolarInvalidTransaction("Transaction could not be verified")

        return True

    def verify_secondsig(self, secondPublicKey):
        msg = self.to_bytes(False, True)

        if self.version > 2:
            # schnorr bip340
            is_valid = verify_schnorr(msg, secondPublicKey, self.signSignature)
        else:
            # schnorr legacy
            is_valid = verify_schnorr_legacy(msg, secondPublicKey, self.signSignature)

        if not is_valid:
            raise SolarInvalidTransaction("Transaction could not be verified")

        return True

    def verify_signatures(self, multi_signature_asset):
        if not multi_signature_asset:
            raise SolarInvalidTransaction("Transaction could not be verified")

        signatures = self.signatures
        if not signatures:
            raise SolarInvalidTransaction("Transaction could not be verified")

        public_keys = multi_signature_asset["publicKeys"]
        min_sigs = multi_signature_asset["min"]

        msg = self.to_bytes(True, True, True)

        public_key_indexes = []
        verified = False
        verified_signatures = 0

        for signature in signatures:
            public_key_index = int(signature[0:2], 16)
            if public_key_index in public_key_indexes:
                raise SolarInvalidTransaction("Transaction could not be verified")
            else:
                public_key_indexes.append(public_key_index)

            partial_signature = signature[2:130]
            public_key = public_keys[public_key_index]
            if self.version > 2:
                # schnorr bip340
                msg_hash = msg
                is_valid = verify_schnorr(msg_hash, public_key, partial_signature)
            else:
                # schnorr legacy
                is_valid = verify_schnorr_legacy(msg, public_key, partial_signature)

            if is_valid:
                verified_signatures += 1

            if verified_signatures >= min_sigs:
                verified = True
                break

        return verified

    def _handle_transaction_type(self, bytes_data):
        """Handled each transaction type differently

        Args:
            bytes_data (bytes): input the bytes data to which you want to append new bytes

        Raises:
            NotImplementedError: raised only if the child transaction doesn't implement this
            required method
        """
        if self.type == TRANSACTION_SECOND_SIGNATURE_REGISTRATION:
            public_key = self.asset["signature"]["publicKey"]
            bytes_data += unhexlify(public_key)
        elif self.type == TRANSACTION_DELEGATE_REGISTRATION:
            bytes_data += self.asset["delegate"]["username"].encode()
        elif self.type == TRANSACTION_VOTE:
            bytes_data += "".join(self.asset["votes"]).encode()
        elif self.type == TRANSACTION_MULTI_SIGNATURE_REGISTRATION:
            bytes_data += write_bit8(self.asset["multiSignature"]["min"])
            bytes_data += "".join(self.asset["multiSignature"]["publicKeys"]).encode()

        return bytes_data

    def _handle_signature(
        self, bytes_data, skip_signature, skip_second_signature, skip_multi_signature
    ):
        """Internal method, used to handle the signature

        Args:
            bytes_data (bytes): input the bytes data to which you want to append new bytes from
            signature
            skip_signature (bool): whether you want to skip it or not
            skip_second_signature (bool): whether you want to skip it or not

        Returns:
            bytes: bytes string
        """
        if not skip_signature and self.signature:
            bytes_data += write_high(self.signature)
        if not skip_second_signature and self.signSignature:
            bytes_data += write_high(self.signSignature)
        if not skip_multi_signature and self.signatures:
            bytes_data += write_high(self.signatures)
        return bytes_data



from binascii import hexlify, unhexlify

from base58 import b58encode_check
from binary.unsigned_integer.reader import read_bit16, read_bit64

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class MultiPaymentDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset / 2)

        payment_length = read_bit16(self.serialized, starting_position) & 0xFF

        self.transaction.asset["payments"] = []

        index = 0

        for payment in range(payment_length):
            amount = read_bit64(self.serialized, offset=starting_position + 2 + index)

            recipient_start_index = (starting_position + 10 + index) * 2
            recipientId = hexlify(self.serialized)[
                recipient_start_index : recipient_start_index + 42
            ]
            recipientId = b58encode_check(unhexlify(recipientId)).decode()

            self.transaction.asset["payments"].append(
                {"amount": amount, "recipientId": recipientId}
            )

            index += 21 + 8

        self.transaction.parse_signatures(
            hexlify(self.serialized).decode(),
            self.asset_offset + 4 + (payment_length * (21 + 8)) * 2,
        )

        return self.transaction



class SolarCryptoException(Exception):
    pass


class SolarSerializerException(SolarCryptoException):
    """Raised when there's a serializer related issue"""


class SolarDeserializerException(SolarCryptoException):
    """Raised when there's a deserializer related issue"""


class SolarInvalidTransaction(SolarCryptoException):
    """Raised when transaction is not valid"""



from binascii import hexlify, unhexlify

from base58 import b58decode_check
from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit8, write_bit32, write_bit64

from solar_crypto.exceptions import SolarSerializerException
from solar_crypto.identity import address
from solar_crypto.transactions.serializers.base import BaseSerializer


class HtlcLockSerializer(BaseSerializer):
    """Serializer handling timelock transfer data"""

    def serialize(self):
        if not address.validate_address(self.transaction["recipientId"]):
            raise SolarSerializerException("Invalid recipient address")

        secret_hash = unhexlify(self.transaction["asset"]["lock"]["secretHash"].encode())

        self.bytes_data += write_bit64(self.transaction["amount"])
        self.bytes_data += write_bit8(len(secret_hash))
        self.bytes_data += secret_hash
        self.bytes_data += write_bit8(self.transaction["asset"]["lock"]["expiration"]["type"])
        self.bytes_data += write_bit32(self.transaction["asset"]["lock"]["expiration"]["value"])

        recipientId = hexlify(b58decode_check(self.transaction["recipientId"]))
        self.bytes_data += write_high(recipientId)

        return self.bytes_data



from binascii import unhexlify

from binary.unsigned_integer.writer import write_bit8

from solar_crypto.transactions.serializers.base import BaseSerializer


class HtlcClaimSerializer(BaseSerializer):
    """Serializer handling timelock claim data"""

    def serialize(self):
        unlock_secret = unhexlify(self.transaction["asset"]["claim"]["unlockSecret"].encode())
        self.bytes_data += write_bit8(self.transaction["asset"]["claim"]["hashType"])
        self.bytes_data += unhexlify(self.transaction["asset"]["claim"]["lockTransactionId"])
        self.bytes_data += write_bit8(len(unlock_secret))
        self.bytes_data += unlock_secret
        return self.bytes_data



from solar_crypto.constants import SOLAR_TRANSACTION_BURN, TRANSACTION_TYPE_GROUP
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class Burn(BaseTransactionBuilder):

    transaction_type = SOLAR_TRANSACTION_BURN
    typeGroup = TRANSACTION_TYPE_GROUP.SOLAR.value

    def __init__(self, amount):
        """Create a burn transaction

        Args:
            amount (int): amount of coins you want to burn
        """
        super().__init__()

        if type(amount) == int and amount > 0:
            self.transaction.amount = amount
        else:
            raise ValueError("Amount is not valid")



from binascii import hexlify

from solar_crypto.transactions.deserializers.base import BaseDeserializer


class SecondSignatureRegistrationDeserializer(BaseDeserializer):
    def deserialize(self):
        starting_position = int(self.asset_offset)

        public_key = hexlify(self.serialized)[starting_position : starting_position + 66]

        self.transaction.asset = {"signature": {"publicKey": public_key.decode()}}

        self.transaction.parse_signatures(hexlify(self.serialized).decode(), self.asset_offset + 66)

        return self.transaction



import sys

import setuptools


# Utility function to read the README file.
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requires = ["base58", "binary-helpers", "coincurve", "btclib", "pycryptodome"]

tests_require = [
    "flake8==4.0.1",
    "flake8-import-order==0.18.1",
    "flake8-print==4.0.1",
    "flake8-quotes==3.3.1",
    "pytest==7.0.1; python_version < '3.7'",
    "pytest==7.1.2; python_version >= '3.7'",
    "isort==5.10.1",
    "pytest-cov==2.5.1",
    "black==22.3.0",
    "build==0.8.0",
    "twine==3.8.0",
]

extras_require = {"test": tests_require, "dev": requires + tests_require}

setup_requires = ["pytest-runner"] if {"pytest", "test", "ptr"}.intersection(sys.argv) else []

setuptools.setup(
    name="solar-crypto",
    description="A simple Cryptography Implementation in Python for the Solar Blockchain.",
    version="3.0.0",
    author="Solar-Network",
    author_email="hello@solar.org",
    url="https://github.com/Solar-network/python-crypto",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requires,
    extras_require=extras_require,
    tests_require=tests_require,
    setup_requires=setup_requires,
    python_requires=">=3.6",
)



from binascii import unhexlify

from solar_crypto.transactions.serializers.base import BaseSerializer


class SecondSignatureRegistrationSerializer(BaseSerializer):
    """Serializer handling registration of the seconds signature"""

    def serialize(self):
        self.bytes_data += unhexlify(self.transaction["asset"]["signature"]["publicKey"].encode())
        return self.bytes_data



from binascii import hexlify

from base58 import b58decode_check
from binary.hex.writer import write_high
from binary.unsigned_integer.writer import write_bit32, write_bit64

from solar_crypto.exceptions import SolarSerializerException
from solar_crypto.identity import address
from solar_crypto.transactions.serializers.base import BaseSerializer


class TransferSerializer(BaseSerializer):
    """Serializer handling transfer data"""

    def serialize(self):
        if not address.validate_address(self.transaction["recipientId"]):
            raise SolarSerializerException("Invalid recipient address")

        self.bytes_data += write_bit64(self.transaction["amount"])
        self.bytes_data += write_bit32(self.transaction.get("expiration", 0))
        recipientId = hexlify(b58decode_check(self.transaction["recipientId"]))
        self.bytes_data += write_high(recipientId)

        return self.bytes_data



import typing

from solar_crypto.constants import TRANSACTION_VOTE
from solar_crypto.identity.address import address_from_passphrase
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class Vote(BaseTransactionBuilder):

    transaction_type = TRANSACTION_VOTE

    def __init__(self, vote=None, fee=None):
        """Create a second signature registration transaction

        Args:
            vote (str): address of a delegate you want to vote
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        self.transaction.asset["votes"] = []
        if vote:
            self.transaction.asset["votes"].append(vote)

        if fee:
            self.transaction.fee = fee

    def set_votes(self, votes: typing.List[str]):
        """Set votes/unvotes

        Args:
            votes (List[str]): list of votes
        """
        self.transaction.asset["votes"] = votes

    def sign(self, passphrase):
        self.transaction.recipientId = address_from_passphrase(passphrase)
        super().sign(passphrase)



from solar_crypto.constants import TRANSACTION_HTLC_CLAIM, TRANSACTION_TYPE_GROUP, HashingType
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class HtlcClaim(BaseTransactionBuilder):

    transaction_type = TRANSACTION_HTLC_CLAIM

    def __init__(
        self,
        lock_transaction_id: str,
        unlock_secret: str,
        hash_type: HashingType = HashingType.SHA256,
        fee: int = None,
    ):
        """Create a timelock transaction

        Args:
            lock_transaction_id (str): HTLC lock transaction ID we wish to claim
            unlock_secret (str): unlock secret required to claim the transaction, must be 64 chars long
            hash_type (HashingType): which hashing algorithm to use to verify the secret
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        self.transaction.typeGroup = self.get_type_group()

        self.transaction.asset["claim"] = {
            "hashType": hash_type.value,
            "lockTransactionId": lock_transaction_id,
            "unlockSecret": unlock_secret,
        }

        if fee:
            self.transaction.fee = fee

    def get_type_group(self):
        return TRANSACTION_TYPE_GROUP.CORE.value



from solar_crypto.constants import TRANSACTION_FEES, TRANSACTION_MULTI_SIGNATURE_REGISTRATION
from solar_crypto.transactions.builder.base import BaseTransactionBuilder


class MultiSignatureRegistration(BaseTransactionBuilder):

    transaction_type = TRANSACTION_MULTI_SIGNATURE_REGISTRATION

    def __init__(self, fee=None):
        """Create a new multi signature transaction

        Args:
            fee (int, optional): fee used for the transaction (default is already set)
        """
        super().__init__()

        self.transaction.asset = {
            "multiSignature": {
                "min": None,
                "publicKeys": [],
            },
        }

        if fee:
            self.transaction.fee = fee

    def set_min(self, minimum_participants):
        self.transaction.asset["multiSignature"]["min"] = minimum_participants

    def set_public_keys(self, public_keys):
        self.transaction.asset["multiSignature"]["publicKeys"] = public_keys
        self.transaction.fee = (len(public_keys) + 1) * self.transaction.fee

    def add_participant(self, public_key):
        self.transaction.asset["multiSignature"]["publicKeys"].append(public_key)
        self.transaction.fee = (
            len(self.transaction.asset["multiSignature"]["publicKeys"]) + 1
        ) * TRANSACTION_FEES.get(TRANSACTION_MULTI_SIGNATURE_REGISTRATION)
